from typing import Type

from monosi.monitors.base import Monitor, MonitorType
from monosi.monitors.custom import CustomMonitor
from monosi.monitors.table import TableMonitor

def load_monitor_cls(monitor_dict) -> Type[Monitor]:
    type_raw = monitor_dict.get('type')
    monitor_type = MonitorType(type_raw)

    if monitor_type == MonitorType.TABLE:
        return TableMonitor
    elif monitor_type == MonitorType.CUSTOM:
        return CustomMonitor

    # Note: Unreachable - we would error at MonitorType instantiation
    raise Exception("Could not find a moniotr with type: {}".format(type_raw))



import os
from flask import Flask
from flask_cors import CORS

from monosi.scheduler.manager import JobManager
from monosi.tasks.schedule import ScheduleMonitorsTask

def load_monitors(app):
    manager = JobManager(app)
    args = None
    schedule_task = ScheduleMonitorsTask.from_args(args, manager)
    schedule_task.run()

def create_app():
    app = Flask(__name__)
    CORS(app, resources={r"/v1/api/*": {"origins": "*"}})

    app_settings = os.getenv(
        'APP_SETTINGS',
        'monosi.server.config.Config'
    )
    app.config.from_object(app_settings)
    
    load_monitors(app)

    return app




from typing import Iterable

from monosi.monitors import load_monitor_cls
from monosi.monitors.base import Monitor
from monosi.parsers import YamlParser
from monosi.utils.files import File

class MonitorParser(YamlParser):
    def __init__(self, configuration):
        super().__init__(configuration, ['monosi', 'monitors'])

    def _fqtablename(self, table):
        table_parts = table.split('.')
        
        if len(table_parts) > 3:
            raise Exception("Can't resolve table name: {}".format(table))
        
        if len(table_parts) == 1 and self.default_schema:
            table_parts.insert(0, self.default_schema)
        if len(table_parts) == 2 and self.default_database:
            table_parts.insert(0, self.default_database)

        return ".".join(table_parts)

    def _resolve_table(self, monitor_dict):
        if 'type' not in monitor_dict or monitor_dict['type'] != 'table':
            return

        fqn = self._fqtablename(monitor_dict['table'])
        monitor_dict['table'] = fqn

    def parse_monitors(self, file: File) -> Iterable[Monitor]:
        monitors_dict = self.extract_from_file(file)

        for monitor_dict in monitors_dict:
            try:
                monitor_cls = load_monitor_cls(monitor_dict)
                monitor_cls.validate(monitor_dict)
                self._resolve_table(monitor_dict)
                monitor = monitor_cls.from_dict(monitor_dict)
            except Exception as e:
                raise e
            yield monitor

    def parse_file(self, file, project):
        for monitor in self.parse_monitors(file):
            project.add_monitor(monitor)




from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List
from monosi.drivers.column import ColumnDataType

from monosi.drivers.column import Column

from .base import Monitor
from .metrics import MetricBase

def extract_or_default(obj, key, default):
    return obj[key] if key in obj else default

class ColumnMetricType(Enum):
    APPROX_DISTINCTNESS = 'approx_distinctness'
    COMPLETENESS = 'completeness'
    ZERO_RATE = 'zero_rate'
    NEGATIVE_RATE = 'negative_rate'
    NUMERIC_MEAN = 'numeric_mean'
    NUMERIC_MIN = 'numeric_min'
    NUMERIC_MAX = 'numeric_max'
    NUMERIC_STD = 'numeric_std'
    APPROX_DISTINCT_COUNT = 'approx_distinct_count'
    MEAN_LENGTH = 'mean_length'
    MAX_LENGTH = 'max_length'
    MIN_LENGTH = 'min_length'
    STD_LENGTH = 'std_length'
    TEXT_INT_RATE = 'text_int_rate'
    TEXT_NUMBER_RATE = 'text_number_rate'
    TEXT_UUID_RATE = 'text_uuid_rate'
    TEXT_ALL_SPACES_RATE = 'text_all_spaces_rate'
    TEXT_NULL_KEYWORD_RATE = 'text_null_keyword_rate'

    @classmethod
    def default(cls):
        return [cls.APPROX_DISTINCTNESS, cls.COMPLETENESS]

    @classmethod
    def all(cls):
        return [
            cls.APPROX_DISTINCTNESS, 
            cls.COMPLETENESS,
            cls.ZERO_RATE,
            cls.NEGATIVE_RATE,
            cls.NUMERIC_MEAN,
            cls.NUMERIC_MIN,
            cls.NUMERIC_MAX,
            cls.NUMERIC_STD,
            cls.APPROX_DISTINCT_COUNT,
            cls.MEAN_LENGTH,
            cls.MAX_LENGTH,
            cls.MIN_LENGTH,
            cls.STD_LENGTH,
            cls.TEXT_INT_RATE,
            cls.TEXT_NUMBER_RATE,
            cls.TEXT_UUID_RATE,
            cls.TEXT_ALL_SPACES_RATE,
            cls.TEXT_NULL_KEYWORD_RATE,
        ]

    @classmethod
    def default_for(cls, data_type: ColumnDataType): 
        if data_type == ColumnDataType.FLOAT or data_type == ColumnDataType.INTEGER:
            return [
                cls.ZERO_RATE,
                cls.NEGATIVE_RATE,
                cls.NUMERIC_MEAN,
                cls.NUMERIC_MIN,
                cls.NUMERIC_MAX,
                cls.NUMERIC_STD,
            ]
        elif data_type == ColumnDataType.STRING:
            return [
                cls.APPROX_DISTINCT_COUNT,
                cls.MEAN_LENGTH,
                cls.MAX_LENGTH,
                cls.MIN_LENGTH,
                cls.STD_LENGTH,
                cls.TEXT_INT_RATE,
                cls.TEXT_NUMBER_RATE,
                cls.TEXT_UUID_RATE,
                cls.TEXT_ALL_SPACES_RATE,
                cls.TEXT_NULL_KEYWORD_RATE,
            ]
            

        return cls.default()


@dataclass
class ColumnMetric(MetricBase):
    type: ColumnMetricType
    column: str

    def alias(self):
        return "{}__{}".format(self.column, self.type.value)

@dataclass 
class TableMonitor(Monitor):
    table: str = field(default_factory=str)
    timestamp_field: str = field(default_factory=str)
    where: str = field(default_factory=str)
    days_ago: int = -100
    metrics: List[ColumnMetricType] = field(default_factory=lambda: ColumnMetricType.all())
    columns: List[Column] = field(default_factory=list)

    # TODO: BASE_SQL should be delegated to Dialect
    BASE_SQL = """
    SELECT 
        DATE_TRUNC('HOUR', {timestamp_field}) as window_start, 
        DATEADD(hr, 1, DATE_TRUNC('HOUR', {timestamp_field})) as window_end, 

        COUNT(*) as row_count, 

        {select_sql}

    FROM {table}
    WHERE 
        DATE_TRUNC('HOUR', {timestamp_field}) >= DATEADD(day, {days_ago}, CURRENT_TIMESTAMP()) 
    GROUP BY window_start, window_end 
    ORDER BY window_start ASC;
    """

    def base_sql_statement(self, select_sql):
        return TableMonitor.BASE_SQL.format(
            select_sql=select_sql,
            table=self.table,
            timestamp_field=self.timestamp_field,
            days_ago=self.days_ago,
        )

    def info(self):
        info_str = "Table Health Monitor: {}".format(self.table)
        if self.description:
            info_str += "\n{}".format(self.description)
        return info_str

    @classmethod
    def validate(cls, monitor_dict):
        pass

    def _create_metrics(self):
        metrics = []
        for column in self.columns:
            for metric_type in ColumnMetricType.default_for(column.data_type):
                if metric_type in self.metrics:
                    metric = ColumnMetric(
                        column=column.name,
                        type=ColumnMetricType(metric_type),
                    )
                    metrics.append(metric)

        return metrics
    
    @classmethod
    def from_dict(cls, value: Dict[str, Any]) -> 'Monitor':
        table = value['table']
        timestamp_field = value['timestamp_field']
        description = extract_or_default(value, 'description', None)
        where = extract_or_default(value, 'where', '')
        days_ago = extract_or_default(value, 'days_ago', -100)
        
        return cls(
            table=table,
            description=description,
            timestamp_field=timestamp_field,
            where=where,
            days_ago=days_ago,
        )

    def to_dict(self):
        # Used for output on bootstrap right now.
        return {
            'table': self.table,
            'timestamp_field': self.timestamp_field,
        }

    def retrieve_metrics(self):
        return self._create_metrics()





import abc

class Dialect:
    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'approx_distinct_count') and 
                callable(subclass.approx_distinct_count) and 
                NotImplemented)

    @abc.abstractclassmethod
    def approx_distinct_count(cls):
        raise NotImplementedError

class GenericDialect(Dialect):
    @classmethod
    def approx_distinct_count(cls):
        return "COUNT(DISTINCT {})"
    
    @classmethod
    def approx_distinctness(cls):
        return "{} / CAST(COUNT(*) AS NUMERIC)".format(cls.approx_distinct_count())

    @classmethod
    def numeric_mean(cls):
        return "AVG({})"

    @classmethod
    def numeric_min(cls):
        return "MIN({})"

    @classmethod
    def numeric_max(cls):
        return "MAX({})"

    @classmethod
    def numeric_std(cls):
        return "STDDEV(CAST({} as double))"

    @classmethod
    def mean_length(cls):
        return cls.numeric_mean().format("LENGTH({})")

    @classmethod
    def max_length(cls):
        return cls.numeric_max().format("LENGTH({})")

    @classmethod
    def min_length(cls):
        return cls.numeric_min().format("LENGTH({})")

    @classmethod
    def std_length(cls):
        return cls.numeric_std().format("LENGTH({})")



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, List, Dict

import operator as py_operator
from monosi.monitors.base import Schedule
# from monosi.monitors.base import Schedule

from monosi.monitors.metrics import MetricType

from .base import Monitor
from .metrics import MetricBase

class ComparativeOperatorDefaults:
    EQ = 'eq'
    NE = 'ne'
    GT = 'gt'
    GE = 'ge'
    LT = 'lt'
    LE = 'le'

class PercentOperatorDefaults:
    ABS_INC = 'abs_inc'
    ABS_dec = 'abs_dec'
    REL_INC = 'rel_inc'
    REL_dec = 'rel_dec'

class Operator(ComparativeOperatorDefaults, PercentOperatorDefaults, Enum):
    EQ = 'eq'
    NE = 'ne'
    GT = 'gt'
    GE = 'ge'
    LT = 'lt'
    LE = 'le'
    ABS_INC = 'abs_inc'
    ABS_dec = 'abs_dec'
    REL_INC = 'rel_inc'
    REL_dec = 'rel_dec'

    def fn(self):
        if self._value_ in ComparativeOperatorDefaults.__dict__.values():
            return getattr(py_operator, self._value_)
        elif self._value_ in PercentOperatorDefaults.__dict__.values():
            return py_operator.iadd # TODO: Update
        
        raise Exception("Operator does not exist.")

@dataclass
class Threshold:
    operator: Operator
    value: float

    def evaluate(self, source_operand: float):
        op_fn = self.operator.fn()
        return op_fn(source_operand, self.value)

    @classmethod
    def from_dict(cls, value: Dict[str, Any]) -> 'Threshold':
        return cls(
            operator=Operator(value['operator']),
            value=value['value'],
        )

@dataclass
class CustomMetric(MetricBase):
    sql: str = field(default_factory=str)
    thresholds: List[Threshold] = field(default_factory=list)
    type: MetricType = MetricType.CUSTOM

    @classmethod
    def from_dict(cls, value: Dict[str, Any]) -> 'CustomMetric':
        sql = value['sql']
        thresholds: List[Threshold] = []

        for threshold_dict in value['thresholds']:
            threshold = Threshold.from_dict(threshold_dict)
            thresholds.append(threshold)

        return cls(
            sql=sql, 
            thresholds=thresholds,
        )

def extract_or_default(obj, key, default):
    return obj[key] if key in obj else default

@dataclass
class CustomMonitor(Monitor):
    metrics: List[CustomMetric] = field(default_factory=list)

    def retrieve_metrics(self):
        return self.metrics

    def info(self):
        info_str = "Custom Monitor"
        if self.description:
            info_str += ": {}".format(self.description)
        return info_str

    @classmethod
    def validate(cls, monitor_dict):
        pass # only one CustomMetric allowed per monitor
    
    @classmethod
    def from_dict(cls, value: Dict[str, Any]) -> 'CustomMonitor':
        metric = CustomMetric.from_dict(value)
        description = extract_or_default(value, 'description', None)
        schedule = Schedule.from_dict(extract_or_default(value, 'schedule', {}))

        return cls(
            metrics=[metric],
            description=description,
            schedule=schedule,
        )

    def base_sql_statement(self, select_sql):
        return "WITH custom_metric AS ({select_sql}) SELECT 1 AS custom FROM custom_metric".format(select_sql=select_sql)




from dataclasses import dataclass, field
import os
import glob
from typing import Any, Dict, List, Optional
from monosi.reporter import Reporter

import monosi.utils.yaml as yaml

@dataclass 
class ProjectConfigurationDefaults:
    version: str = '0.0.2.post1' # TODO: Update to automatic
    collection_name: Optional[str] = None
    source_name: Optional[str] = None
    monitor_paths: List[str] = field(default_factory=lambda: ['./monitors'])
    reporter: Reporter = field(default_factory=Reporter)
    
    @classmethod
    def _retrieve_project_config_path(cls, root_path):
        root_path = os.path.normpath(root_path)

        config_path_selector = os.path.join(root_path, 'monosi_project.y*ml')
        matches = glob.glob(config_path_selector)

        if len(matches) != 1:
            raise Exception("No project configuration file found.")

        project_config_path = matches[0]
        return project_config_path

@dataclass
class ProjectConfigurationBase:
    project_name: str
    root_path: str

@dataclass
class ProjectConfiguration(ProjectConfigurationDefaults, ProjectConfigurationBase):
    @classmethod
    def create_new(cls, project_name: str):
        root_path = os.path.abspath(os.getcwd())

        return cls(
            project_name=project_name,
            root_path=root_path,
        )

    @classmethod
    def from_root_path(cls, root_path: str) -> 'ProjectConfiguration':
        project_config_path = cls._retrieve_project_config_path(root_path)
        project_dict = yaml.parse_file(project_config_path)

        return cls.from_dict(
            root_path=root_path,
            project_dict=project_dict,
        )

    @classmethod
    def validate(cls, project_dict):
        pass

    @classmethod
    def from_dict(
        cls,
        root_path: str,
        project_dict: Dict[str, Any],
    ):
        try:
            cls.validate(project_dict)
        except Exception as e:
            raise e

        project_name = str(project_dict.get('name'))
        version = project_dict.get('version') or '0.0.2.post1'
        collection_name = project_dict.get('collection')
        source_name = project_dict.get('source')
        log_path: str = project_dict.get('log-path') or 'logs'

        monitor_paths: List[str] = project_dict.get('monitor-paths') or []

        project = cls(
            project_name=project_name,
            version=version,
            root_path=root_path,
            collection_name=collection_name,
            source_name=source_name,
            monitor_paths=monitor_paths,
        )

        return project

    def to_dict(self):
        config_dict = {
            "name": self.project_name,
            "version": self.version,
            "monitor-paths": self.monitor_paths,
        }
        if self.collection_name:
            config_dict.update({"collection": self.collection_name})
        if self.source_name:
            config_dict.update({"source": self.source_name})

        return config_dict





from monosi.profiler import Profiler
from monosi.tasks.base import TaskBase

class ProfileTask(TaskBase):
    def run(self, *args, **kwargs):
        profiler = Profiler(self.config)
        profiler.profile()
        print("Profiling is complete.")




from dataclasses import dataclass
from typing import Any, List
from monosi.drivers.column import Column
from monosi.profiler import DatabaseTable
from monosi.monitors.table import TableMonitor

from monosi.drivers.dialect import Dialect
from monosi.monitors.base import Monitor
from monosi.monitors.metrics import MetricBase
from monosi.monitors.custom import CustomMetric
from monosi.monitors.table import ColumnMetric, ColumnMetricType

@dataclass
class Compiler:
    dialect: Dialect
    metadata: Any

    def _retrieve_unformatted_sql(self, metric_type: ColumnMetricType) -> str:
        attr = getattr(self.dialect, metric_type._value_)
        if not attr:
            raise Exception("Unreachable: Metric type is defined that does not resolve to a definition.")

        return attr()

    def _compile_column_metric(self, metric: ColumnMetric):
        alias = metric.alias()
        column = metric.column

        unformatted_metric_sql = self._retrieve_unformatted_sql(metric.type)
        formatted_metric_sql = unformatted_metric_sql.format(column)

        metric_sql = "{} AS {}".format(formatted_metric_sql, alias)
        return metric_sql

    def _compile_custom_metric(self, metric: CustomMetric):
        return metric.sql

    def compile_metric(self, metric: MetricBase):
        if isinstance(metric, ColumnMetric):
            return self._compile_column_metric(metric)
        elif isinstance(metric, CustomMetric):
            return self._compile_custom_metric(metric)
        else:
            raise Exception("Metric could not be compiled")

    def compile_select(self, metrics: List[MetricBase]):
        select_body = []
        for metric in metrics:
            metric_sql = self.compile_metric(metric)
            select_body.append(metric_sql)

        return ",\n\t".join(select_body)

    def _add_cols(self, monitor: TableMonitor):
        tables = DatabaseTable.from_metadata(self.metadata)
        for table in tables:
            if table.name.lower() in monitor.table:
                monitor.columns = table.columns

    def compile(self, monitor: Monitor):
        if isinstance(monitor, TableMonitor):
            self._add_cols(monitor)

        select_sql = self.compile_select(monitor.retrieve_metrics())
        sql = monitor.base_sql_statement(select_sql)
        
        return sql




from datetime import datetime
from time import time

from monosi.reporter.formatters.progress import ProgressFormatter

from .notifications import NotificationType

class Reporter:
    def __init__(self):
        self.tests = []
        self.skipped_tests = []
        self.failed_tests = []
        self.listeners = {}
        self.start_time = None
        self.stop_time = None
        self.monitor_finished_time = None

        formatter = ProgressFormatter()
        for notification_type in NotificationType.all():
            self.listeners[notification_type] = [formatter]
        
    def register_listener(self, listener, *notifications):
        for notification in notifications:
            self.listeners[notification].append(listener)

    def _listeners(self, notification):
        return list(self.listeners[notification])

    def monitor_started(self, monitor):
        self.start()
        self.notify(NotificationType.MONITOR_STARTED, monitor)

    def _reset(self):
        self.tests = []
        self.failed_tests = []
        self.load_time = None
        self.start_time = None
        self.stop_time = None
        self.monitor_finished_time = None

    def monitor_finished(self, monitor):
        self.monitor_finished_time = time()
        self.notify(NotificationType.MONITOR_FINISHED)
        self.finish()

    def test_started(self, test):
        self.tests.append(test)
        self.notify(NotificationType.TEST_STARTED)

    def test_finished(self, test):
        self.notify(NotificationType.TEST_FINISHED)

    def test_passed(self, test):
        self.notify(NotificationType.TEST_PASSED)

    def test_failed(self, test):
        self.failed_tests.append(test)
        self.notify(NotificationType.TEST_FAILED, test)

    def notify_non_test_exception(self, exception, context_description):
        pass

    def start(self):
        self._reset()
        self.start_time = time()

    def stop(self):
        self.stop_time = time()
        self.notify(NotificationType.STOP)

    def notify(self, notification_type: NotificationType, obj=None):
        # ensure listeners are ready
        try:
            [getattr(formatter, notification_type.value)(obj) for formatter in self._listeners(notification_type)]
        except Exception as e:
            print(e)


    def _load_time(self):
        if self.monitor_finished_time and self.start_time:
            return round(self.monitor_finished_time - self.start_time, 2)
        else:
            raise Exception("Could not calculate the load time.")

    def _total_time(self):
        if self.stop_time and self.start_time:
            return round(self.stop_time - self.start_time, 2)
        else:
            raise Exception("Could not calculate the total time.")

    def finish(self):
        self.stop()
        
        summary = {
            'failed_count': len(self.failed_tests),
            'skipped_count': 0,
            'test_count': len(self.tests),
            'total_time': self._total_time(),
            'load_time': self._load_time(),
        }

        self.notify(NotificationType.START_DUMP)
        self.notify(NotificationType.DUMP_SKIPPED)
        self.notify(NotificationType.DUMP_FAILURES, self.failed_tests)
        self.notify(NotificationType.DUMP_SUMMARY, summary)




from snowplow_tracker import Tracker, AsyncEmitter, SelfDescribingJson, Subject, logger as snowplow_logger
import logging
import platform

snowplow_logger.setLevel(100)
SNOWPLOW_URL = "8345dd55-fcc1-402e-a7ef-cf044d2b8f02.app.try-snowplow.com" 
# check if dev
e = AsyncEmitter(
    SNOWPLOW_URL, 
    protocol="https",
)
tracker = Tracker(
    e, 
    namespace="cf", 
    app_id="monosi",
)

def set_user_id(user_id):
    subject = Subject()
    subject.set_user_id(user_id)
    tracker.set_subject(subject)

def context():
    return [
        SelfDescribingJson(
            schema="iglu:com.trysnowplow/object/jsonschema/1-0-0",
            data={
                "name": "platform_context",
                "svalue": platform.platform(),
                # "svalue": platform.python_version(),
                # "svalue": platform.python_implementation(),
            },
        )
    ]

def track_event(config, *args, **kwargs):
    if config.send_anonymous_stats:
        try:
            tracker.track_struct_event(
                action=kwargs['action'],
                label=kwargs['label'],
                category='monosi',
                context=context(),
            )
        except Exception as e:
            logging.error("Failed to send anonymous usage stats.")
            logging.error(e)




from dataclasses import dataclass
from math import sqrt
from typing import List
from monosi.analyzer.data import DataPoint

from monosi.monitors.table import ColumnMetric

from .data import Data, Test, TestResult, TableDataPoint

@dataclass
class ZScoreDataPointFields:
    expected_range_start: float
    expected_range_end: float
    z_score: float

@dataclass
class ZScoreDataPoint(TableDataPoint, ZScoreDataPointFields):
    pass

class ZScoreAlgorithm:
    @classmethod
    def _mean(cls, values: List[float]):
        return round(sum(values) / len(values), 2)

    @classmethod
    def _std_dev(cls, values: List[float]):
        values_mean = cls._mean(values)

        distances = [((value - values_mean) ** 2) for value in values]
        distances_mean = cls._mean(distances)
        std_dev = round(sqrt(distances_mean), 2)

        return std_dev

    @classmethod
    def run(cls, points: List[DataPoint], sensitivity: float):
        nonnull_points = list(filter(lambda x: x.value != None, points))
        values = [point.value for point in nonnull_points]
        try:
            mean = cls._mean(values)
            std_dev = cls._std_dev(values)
        except ZeroDivisionError:
            return []

        if (std_dev == 0): return []

        zscore_points = []
        for point in nonnull_points:
            try:
                z_score = round(((point.value - mean) / std_dev), 2) 
                error = abs(z_score) > sensitivity 
                zscore_point = ZScoreDataPoint(
                    **point.to_dict(),
                    expected_range_start=point.value-sensitivity,
                    expected_range_end=point.value+sensitivity,
                    error=error,
                    z_score=z_score,
                )
                zscore_points.append(zscore_point)

            except TypeError:
                return []

        return zscore_points

class ZScoreTestResult(TestResult):
    data: List[ZScoreDataPoint]

@dataclass
class ZScoreTest(Test):
    sensitivity: float = 3.0
    data: List[DataPoint]

    @classmethod
    def from_metric(cls, metric: ColumnMetric, data: Data):
        metric_data = data.for_metric(metric)
        return cls(data=metric_data)

    def run(self):
        z_scores = ZScoreAlgorithm.run(self.data, self.sensitivity)

        return ZScoreTestResult(z_scores)




from dataclasses import dataclass
from typing import Any, Tuple
import os

import monosi.utils.yaml as yaml

from .project import ProjectConfiguration, ProjectConfigurationBase, ProjectConfigurationDefaults
from .collection import CollectionConfiguration, CollectionConfigurationBase, CollectionConfigurationDefaults

@dataclass
class ConfigurationBase(ProjectConfigurationBase, CollectionConfigurationBase):
    pass

@dataclass
class ConfigurationDefaults(ProjectConfigurationDefaults, CollectionConfigurationDefaults):
    args: Any = None

@dataclass
class Configuration(ConfigurationDefaults, ConfigurationBase):
    def __post_init__(self):
        try:
            Configuration.validate(self)
        except Exception as e:
            raise e

    @classmethod
    def validate(cls, self):
        pass

    @classmethod
    def create_subclasses(cls, args: Any) -> Tuple[ProjectConfiguration, CollectionConfiguration]:
        root_path = os.getcwd()
        if hasattr(args, 'project_dir'):
            root_path = args.project_dir

        project = ProjectConfiguration.from_root_path(root_path)

        collection_name = (project.collection_name or 'default')
        source_name = (project.source_name or 'default')
        collection = CollectionConfiguration.from_args(
            collection_name=collection_name, 
            source_name=source_name,
            args=args
        )

        return (project, collection)

    @classmethod
    def from_subclasses(cls, project: ProjectConfiguration, collection: CollectionConfiguration, args) -> 'Configuration':
        return cls(
            project_name=project.project_name,
            version=project.version,
            root_path=project.root_path,
            monitor_paths=project.monitor_paths,
            collection_name=project.collection_name,
            source_name=project.source_name,
            reporter=project.reporter,
            config=collection.config,
            send_anonymous_stats=collection.send_anonymous_stats,
            args=args)

    @classmethod
    def from_args(cls, args: Any) -> 'Configuration':
        project, collection = cls.create_subclasses(args)
        configuration = cls.from_subclasses(
            project=project,
            collection=collection,
            args=args
        )

        return configuration

    def project_dict(self):
        config_dict = {
            "name": self.project_name,
            "version": self.version,
            "monitor-paths": self.monitor_paths,
        }
        if self.collection_name:
            config_dict.update({"collection": self.collection_name})
        if self.source_name:
            config_dict.update({"source": self.source_name})

        return config_dict

    def add_monitor_path(self, path):
        if path not in self.monitor_paths:
            self.monitor_paths.append(path)
            
            project_config_path = self._retrieve_project_config_path(self.root_path)
            project_dict = self.project_dict()

            yaml.write_file(project_config_path, project_dict)



from monosi.events import track_event

class BaseJob:
    @classmethod
    def run_job(cls, *args, **kwargs):
        job = cls()
        return job.run(*args, **kwargs)

    def run(self, *args, **kwargs):
        raise NotImplementedError('Job not implemented.')

class MonitorJob(BaseJob):
    def __init__(self, task):
        self.task = task

    def run(self, *args, **kwargs):
        track_event(self.task.config, action="run_start", label="scheduled")
        self.task.run()
        track_event(self.task.config, action="run_finish", label="1")




import logging

from monosi.monitors.custom import CustomMetric
from monosi.monitors.table import ColumnMetric

from .data import Data, DataPoint, TableData
from .threshold import ThresholdTest
from .zscore import ZScoreTest

class Analyzer:
    def __init__(self, reporter):
        self.reporter = reporter

    def _create_test(self, metric, data: Data):
        if isinstance(metric, ColumnMetric):
            return ZScoreTest.from_metric(metric, data)
        elif isinstance(metric, CustomMetric):
            return ThresholdTest.from_metric(metric, data)
        else:
            logging.warning("Data returned from the database did not contain data for a metric defined in the monitor.")
            raise Exception("Could not create test because the monitor was not recognized.")

    def test(self, metric, data):
        test = self._create_test(metric, data)
        
        try:
            self.reporter.test_started(self)
            result = test.run()
            if not len(result.anomalies()) > 0:
                self.reporter.test_passed(result)
            else:
                self.reporter.test_failed(result)
        finally:
            self.reporter.test_finished(metric)
        
        return result

    def analyze(self, monitor, results):
        data = Data.from_results(results)
        results = {}

        for metric in monitor.retrieve_metrics():
            result = self.test(metric, data)
            results[metric.alias()] = result

        return results




from dataclasses import dataclass
from typing import Dict, List

from monosi.monitors.metrics import MetricBase

@dataclass
class DataPoint:
    value: float
    error: bool = False

    def to_dict(self):
        return {
            'value': self.value,
            'error': self.error,
        }

class TableDataPoint(DataPoint):
    window_start: str
    window_end: str

@dataclass
class TestResult:
    data: List[DataPoint]

    def anomalies(self):
        return list(filter(lambda x: x.error == True, self.data))

@dataclass
class Data:
    points: Dict[str, List[DataPoint]]

    def for_metric(self, metric: MetricBase) :
        alias = metric.alias()
        return self.points[alias.lower()]

    @classmethod
    def anomalies(cls, points: List[DataPoint]):
        raise NotImplementedError

    @classmethod
    def from_results(cls, results):
        cols = [column.name.lower() for column in results['columns']]
        points = {}
        for col in cols:
            points[col] = []

        for row in results['rows']:
            for col in row.keys():
                if row[col]:
                    point = DataPoint(value=float(row[col]))
                    points[col.lower()].append(point)
        
        return cls(points)

@dataclass
class TableData(Data):
    points: Dict[str, List[TableDataPoint]]

    @classmethod
    def from_results(cls, results):
        return super().from_results(results)

    @classmethod
    def anomalies(cls, points: List[DataPoint]):
        pass

class CustomData(Data):
    @classmethod
    def anomalies(cls, points: List[DataPoint]):
        pass

@dataclass
class Test:
    data: List[DataPoint]




import pkg_resources

from monosi.drivers import DriverConfig
import monosi.drivers.snowflake

class DriverFactory:
    def __init__(self):
        self._drivers = {}

        self._drivers['snowflake'] = monosi.drivers.snowflake
        # for entry_point in pkg_resources.iter_entry_points('monosi_drivers'):
        #     self._drivers[entry_point.name] = entry_point.load()

    def load_config_class(self, name: str):
        if name not in self._drivers.keys():
            raise Exception

        return self._retrieve_driver_class(name, 'DriverConfig')

    def load_driver_class(self, config: DriverConfig):
        name = config.driver_name()
        connection_cls = self._retrieve_driver_class(name, 'Driver')

        return connection_cls

    def _configuration(self, config_vals):
        driver_name = config_vals['driver']

        configuration_class = self._retrieve_driver_class(driver_name, 'DriverConfig')
        return configuration_class(**config_vals)

    def _retrieve_driver_class(self, driver_name, class_name):
        module = self._retrieve_driver_module(driver_name)
        return getattr(module, class_name)

    def _retrieve_driver_module(self, name):
        module = self._drivers.get(name)
        if not module:
            raise ValueError(name)
        return module


FACTORY: DriverFactory = DriverFactory()

def load_config(name: str):
    return FACTORY.load_config_class(name)

def load_driver(config: DriverConfig):
    return FACTORY.load_driver_class(config)




from dataclasses import dataclass
from typing import List

import os
import logging
from monosi.drivers.column import ColumnDataType

from monosi.drivers.column import Column
from monosi.monitors.table import TableMonitor
import monosi.utils.yaml as yaml


def resolve_to_type_from_str(type_str):
    type_str = type_str.lower()
    if 'varchar' in type_str:
        return ColumnDataType.STRING
    elif 'number' in type_str:
        return ColumnDataType.INTEGER
    elif 'date' in type_str or 'timestamp' in type_str:
        return ColumnDataType.DATE

    return ColumnDataType.STRING

@dataclass
class DatabaseTable:
    name: str
    columns: List[Column]

    def timestamp_cols(self):
        t_col_types = [ColumnDataType.DATE, ColumnDataType.DATETIME]
        cols = list(filter(lambda x: x.data_type in t_col_types, self.columns))

        return cols

    def timestamp(self):
        return self.timestamp_cols()[0]

    @classmethod
    def _insert_in_table(cls, tables, name, column):
        if name not in tables:
            table = cls(
                name=name,
                columns=[]
            )
        else:
            table = tables[name]

        table.columns.append(column)
        tables[name] = table

        return tables

    @classmethod
    def from_metadata(cls, metadata):
        tables = {}
        for row in metadata['rows']:
            column = Column(
                name=row['COL_NAME'], 
                data_type=resolve_to_type_from_str(row['COL_TYPE']),
            )
            tables = cls._insert_in_table(tables, row['NAME'], column)

        return tables.values()

BOOTSTRAPPED_MONITOR_PATH = './bootstrapped-monitors'
class Profiler:
    def __init__(self, config):
        self.config = config

    def _retrieve_tables(self):
        from monosi.drivers.factory import load_driver
        driver_cls = load_driver(self.config.config)
        driver = driver_cls(self.config)

        metadata = driver.metadata()
        return DatabaseTable.from_metadata(metadata)

    def _create_definitions(self):
        definitions = []

        tables = self._retrieve_tables()
        for table in tables:
            try:
                monitor = TableMonitor(
                    table=table.name,
                    timestamp_field=table.timestamp().name,
                )
                definition = {
                    'monosi': {
                        'monitors': [monitor.to_dict()]
                    }
                }
                definitions.append(definition)
            except:
                pass
                # logging.warn("Timestamp field could not be found for table")

        return definitions

    def _write_definition(self, definition, monitors_dir=BOOTSTRAPPED_MONITOR_PATH):
        path = os.path.join(monitors_dir, definition['monosi']['monitors'][0]['table'] + ".yaml")
        if not os.path.exists(path):
            yaml.write_file(path, definition)

    def _persist_definitions(self, definitions):
        if not os.path.exists(BOOTSTRAPPED_MONITOR_PATH):
            os.makedirs(BOOTSTRAPPED_MONITOR_PATH)
        self.config.add_monitor_path(BOOTSTRAPPED_MONITOR_PATH)

        for definition in definitions:
            self._write_definition(definition)

    def profile(self):
        definitions = self._create_definitions()
        # TODO: validate definitions
        self._persist_definitions(definitions)





from .base import BaseFormatter, Color

class TextFormatter(BaseFormatter):
    def message(self, notification):
        self.write('message', Color.ENDC)

    def dump_failures(self, failed_tests):
        # TODO: Consider Failures Formatter
        if len(failed_tests) == 0:
            self.write('', Color.ENDC)
            # self.write("\n\nAll tests passed.\n", Color.GREEN)
        else:
            self.write("\nFailures\n", Color.RED)

        for failed_test in failed_tests:
            self.write("\tColumn: {}\n".format(failed_test.column), Color.RED)
            self.write("\tMetric: {}\n".format(failed_test.metric), Color.RED)
            self.write("\tAnomalies: {}\n".format(len(failed_test.anomalies)), Color.RED)
            self.write("\n", Color.RED)
            # for anomaly in failed_test.anomalies:
            #     self.write("\tValue: {value}\n\tExpected Range: {start} - {stop}\n\tTime Window: {time_start} - {time_end}\n\n".format(
            #         value=anomaly.point.value,
            #         start=anomaly.expected_range_start,
            #         stop=anomaly.expected_range_end,
            #         time_start=anomaly.point.window_start,
            #         time_end=anomaly.point.window_end,
            #     ), Color.RED)

        # if len(notification.failure_notifications) == 0: 
        #     return


    def dump_summary(self, summary):
        # print(summary.fully_formatted)
        self.write('\n\nFinished in {total_time} seconds (SQL results took {load_time} seconds to load)\n'.format(
            total_time=summary['total_time'],
            load_time=summary['load_time'],
        ), Color.ENDC)

        color = Color.GREEN
        if summary['failed_count'] > 0:
            color = Color.RED

        self.write('{test_count} metrics, {failed_count} failures\n\n\n'.format(
            test_count=summary['test_count'],
            failed_count=summary['failed_count'],
        ), color)
        # self.write('dump_summary', Color.ENDC)

    def dump_pending(self, notification):
        # if len(notification.pending_examples) == 0: 
        #     return

        # print(notification.fully_formatted_pending_examples)
        # self.write('dump_pending', Color.YELLOW)
        pass

    def close(self):
        pass



from monosi.events import track_event

from .run import RunMonitorsTask

class TestMonitorsTask(RunMonitorsTask):
    def _process_tasks(self):
        track_event(config, action='run_start', label='manual')
        super()._process_tasks()
        track_event(self.config, action="run_finish", label=str(len(self.monitors)))



from monosi.analyzer import Analyzer
from monosi.config.configuration import Configuration
from monosi.compiler import Compiler
from monosi.events import track_event

class Runner:
    def __init__(self, config: Configuration, monitors):
        self.config = config
        self.monitors = monitors
        self.driver = None

    def _initialize(self):
        try:
            from monosi.drivers.factory import load_driver
            driver_config = self.config.config
            driver_cls = load_driver(driver_config)

            self.driver = driver_cls(self.config)
        except:
            raise Exception("Could not initialize connection to database in runner.")

    def get_compiler(self):
        if self.driver is None:
            raise Exception("Initialize runner before compilation")

        metadata = self.driver.metadata()

        return Compiler(self.driver.dialect, metadata)

    def execute(self, sql: str):
        if self.driver is None:
            raise Exception("Initialize runner before execution.")

        results = self.driver.execute_sql(sql)
        return results

    def run(self):
        self._initialize()

        reporter = self.config.reporter
        compiler = self.get_compiler()
        analyzer = Analyzer(reporter)

        for monitor in self.monitors:
            reporter.monitor_started(monitor)
            try:

                sql_stmt = compiler.compile(monitor)
                results = self.execute(sql_stmt)

                analyzer.analyze(monitor, results)
            finally:
                reporter.monitor_finished(monitor)
                
        # reporter.finish()




from enum import Enum

class NotificationType(Enum):
    MONITOR_STARTED = 'monitor_started'
    MONITOR_FINISHED = 'monitor_finished'
    TEST_STARTED = 'test_started'
    TEST_FINISHED = 'test_finished'
    TEST_PASSED = 'test_passed'
    TEST_FAILED = 'test_failed'
    START_DUMP = 'start_dump'
    DUMP_SKIPPED = 'dump_skipped'
    DUMP_FAILURES = 'dump_failures'
    DUMP_SUMMARY = 'dump_summary'
    STOP = 'stop'

    @classmethod
    def all(cls):
        return cls.__members__.values()

class StartNotification:
    count: int

class FailedTestNotification:
    pass

class SkippedTestNotification:
    pass

class MonitorNotification:
    pass

class MessageNotification:
    pass

class SummaryNotification:
    pass



from dataclasses import dataclass, field
from typing import List

from monosi.config.configuration import Configuration
from monosi.monitors.base import Monitor
from monosi.parsers.monitors import MonitorParser
from monosi.utils.files import FileType, read_project_files

def parser_for_filetype(filetype: FileType):
    if filetype == FileType.MONITOR:
        return MonitorParser
    
    return None

@dataclass
class Project:
    configuration: Configuration
    monitors: List[Monitor] = field(default_factory=list)

    def add_monitor(self, monitor: Monitor):
        self.monitors.append(monitor)

    @classmethod
    def from_configuration(cls, configuration: Configuration) -> 'Project':
        project = Project(configuration)
        project_files = read_project_files(configuration)

        for filetype in project_files.keys():
            parser_cls = parser_for_filetype(filetype)

            if parser_cls:
                parser = parser_cls(configuration)

                files = project_files[filetype]
                for file in files:
                    parser.parse_file(file, project)

        return project






import os
import sys
from argparse import ArgumentParser
from monosi.config.project import ProjectConfiguration

from monosi.tasks.profile import ProfileTask
from monosi.tasks.run import RunMonitorsTask
from monosi.tasks.server import ServerTask
from monosi.utils.yaml import write_file

class CliParser(object):
    """Argument parser for the CLI"""

    def __init__(self):
        self._parser = ArgumentParser()
        self._add_arguments()

    def _add_arguments(self):
        """Adds arguments to parser."""
        self._parser.add_argument(
            'command', type=str
        )
        self._parser.add_argument(
            '-v', '--version',
            action='version', 
            version=self.version(),
            help='Show the version number.'
        )

    def parse(self, argv):
        args = self._parser.parse_args(argv[1:])
        getattr(self, args.command)(argv[1:])
        
    # Commands
    def version(self, args=None):
        return format_program_version(get_installation_info().version, sys.version.split()[0])

    def init(self, args):
        args = None
        project_name = "project"
        project_configuration = ProjectConfiguration.create_new(project_name)

        project_directory = os.path.join(os.getcwd(), project_name)
        os.mkdir(project_directory)

        project_file = os.path.join(project_directory, 'monosi_project.yml')
        write_file(project_file, project_configuration.to_dict())

        print("Successfully initialized monosi project.")

    def profile(self, args):
        args = None
        task = ProfileTask.from_args(args)
        task.run()

    def run(self, args):
        args = None
        task = RunMonitorsTask.from_args(args)
        task.run()

    def server(self, args):
        args = None
        task = ServerTask.from_args(args)
        task.run()

def get_installation_info():
    import pkg_resources

    return pkg_resources.require('monosi')[0]

def format_program_version(pkg_version, python_version):
    return u'monosi {} using Python {}\n'.format(pkg_version, python_version)




import abc
from dataclasses import dataclass, field
from enum import Enum
from mashumaro import DataClassDictMixin
from typing import Any, List, Optional, Dict

from .metrics import MetricBase

class MonitorType(Enum):
    TABLE = 'table'
    CUSTOM = 'custom'

class ScheduleType(Enum):
    INTERVAL = 'interval'

@dataclass
class Schedule(DataClassDictMixin):
    minutes: int = 720
    type: ScheduleType = ScheduleType.INTERVAL

@dataclass
class Monitor:
    metrics: List[MetricBase]
    description: Optional[str] = None
    schedule: Schedule = field(default_factory=Schedule)

    def base_sql_statement(self, select_sql):
        raise NotImplementedError

    def retrieve_metrics(self):
        raise NotImplementedError

    @abc.abstractmethod
    def info(self):
        raise NotImplementedError

    @abc.abstractclassmethod
    def validate(cls, monitor_dict):
        raise NotImplementedError
    
    @abc.abstractclassmethod
    def from_dict(cls, value: Dict[str, Any]) -> 'Monitor':
        raise NotImplementedError



from .base import BaseFormatter, Color

class FailureListFormatter(BaseFormatter):
    def example_failed(self, failure):
        self.write("{location}:{description}".format(
            location=failure.example.location, 
            description=failure.example.description
        ), Color.RED)



from dataclasses import dataclass
from enum import Enum
from os import path
from typing import Optional
import glob

class FileType(Enum):
    MONITOR = 'monitor'
    # PROJECT_CONFIG = 'project_config'
    # COLLECTION_CONFIG = 'collection_config'

@dataclass
class File:
    filepath: str
    name: str
    contents: Optional[str]

    def __init__(self, filepath: str, skip_load=False):
        self.filepath = filepath
        self.name = path.basename(filepath)

        if not skip_load:
            self.load()

    def load(self):
        with open(self.filepath, 'rb') as f:
            self.contents = f.read().decode('utf-8')

def file_search(directory: str, extension: str):
    abs_path = path.abspath(directory)
    regex_str = "{}/**/*.{}".format(abs_path, extension)
    
    filepaths = glob.glob(regex_str, recursive=True)

    return filepaths

def read_files(paths, extension):
    files = []

    for path in paths:
        filepaths = file_search(path, extension)

        for filepath in filepaths:
            file = File(filepath)
            files.append(file)

    return files

def read_project_files(project):
    files = {}
    files[FileType.MONITOR] = read_files(project.monitor_paths, 'y*ml')

    return files




from dataclasses import dataclass
from typing import List

from monosi.monitors.custom import CustomMetric, Threshold

from .data import Data, Test, TestResult

@dataclass
class ThresholdTest(Test):
    thresholds: List[Threshold]

    def run(self):
        for point in self.data:
            for threshold in self.thresholds:
                point.error = threshold.evaluate(point.value)

        return TestResult(self.data)

    @classmethod
    def from_metric(cls, metric: CustomMetric, data: Data):
        return cls(
            data=data.for_metric(metric),
            thresholds=metric.thresholds,
        )




from dataclasses import dataclass
import logging
from typing import Any, Dict, Optional, Tuple
import glob
import os
import uuid

from monosi.drivers import DriverConfig
import monosi.events as events
import monosi.utils.yaml as yaml

DEFAULT_COLLECTIONS_DIR = os.path.join(os.path.expanduser('~'), '.monosi')

def read_user_id(user_filepath: str):
    user_data = yaml.parse_file(user_filepath)
    return user_data['id']

def write_user_id(user_filepath: str):
    user_id = str(uuid.uuid4())

    if not os.path.exists(user_filepath):
        user_data = {"id": user_id}
        yaml.write_file(user_filepath, user_data)

    return user_id

def convert_to_bool(val):
    if val and val.lower() == "true":
        return True
    
    return False

@dataclass 
class CollectionConfigurationDefaults:
    collections_dir: str = DEFAULT_COLLECTIONS_DIR # TODO: Update to args
    send_anonymous_stats: bool = True

@dataclass
class CollectionConfigurationBase:
    config: DriverConfig

@dataclass
class CollectionConfiguration(CollectionConfigurationDefaults, CollectionConfigurationBase):
    @classmethod
    def from_driver_config(
        cls,
        config: DriverConfig,
        send_anonymous_stats: bool,
    ) -> 'CollectionConfiguration':
        collection = cls(
            config=config,
            send_anonymous_stats=send_anonymous_stats,
        )
        collection.validate()
        collection._initialize_events()
        return collection
    
    def _initialize_events(self):
        if not self.send_anonymous_stats:
            return

        user_filepath = os.path.join(self.collections_dir, '.cookie.yml')
        try:
            if os.path.exists(user_filepath):
                user_id = read_user_id(user_filepath)
            else:
                user_id = write_user_id(user_filepath)
            events.set_user_id(user_id)
        except:
            logging.error("There was an issue sending anonymous usage stats with a user id.")

    @staticmethod
    def _config_from_source(source_dict: Dict[str, Any]) -> DriverConfig:
        from monosi.drivers.factory import load_config

        if 'type' not in source_dict:
            raise Exception("Source type is required.")

        driver_type = source_dict.pop('type')
        try:
            cls = load_config(driver_type)
            data = cls.retrieve_data(source_dict)
            cls.validate(data)
            config = cls.from_dict(data)
        except Exception as e:
            raise e

        return config

    @classmethod
    def _get_source_dict(cls, collection_dict: Dict[str, Any], source_name: Optional[str]):
        if 'sources' not in collection_dict:
            raise Exception
        sources = collection_dict['sources']

        if source_name not in sources:
            raise Exception("Source name not found")
        source_dict = sources[source_name]

        if not isinstance(source_dict, dict):
            raise Exception

        return source_dict

    @classmethod
    def _get_collection_dict(cls, collection_name: str, collections_dict: Dict[str, Any]):
        if collection_name not in collections_dict:
            raise Exception("Collection not found: {}".format(collection_name))

        collection_dict = collections_dict[collection_name]
        if not collection_dict:
            raise Exception

        return collection_dict

    def validate(self):
        if not os.path.exists(self.collections_dir):
            raise Exception("Collections directory does not exist.")

        return True

    @classmethod
    def from_dict(cls, collection_dict: Dict[str, Any], source_name: str = 'default') -> 'CollectionConfiguration':
        source_dict = cls._get_source_dict(collection_dict, source_name)
        config = cls._config_from_source(source_dict)

        send_anonymous_stats = convert_to_bool(collection_dict['send_anonymous_stats']) if 'send_anonymous_stats' in collection_dict else True

        return cls.from_driver_config(
            config=config,
            send_anonymous_stats=send_anonymous_stats
        )

    @classmethod
    def _retrieve_collections_path(cls, collections_dir):
        collections_path_selector = os.path.join(collections_dir, 'collections.y*ml')
        matches = glob.glob(collections_path_selector)

        if len(matches) != 1:
            raise Exception("There was no monosi collections configuration file.")

        collections_path = matches[0]
        return collections_path
        
    @classmethod
    def from_args(cls, collection_name='default', source_name='default', args=None) -> 'CollectionConfiguration':
        collections_path = cls._retrieve_collections_path(DEFAULT_COLLECTIONS_DIR)
        collections_dict = yaml.parse_file(collections_path)

        # Should check args for a collection name
        collection_dict = cls._get_collection_dict(collection_name, collections_dict)

        return cls.from_dict(
            collection_dict=collection_dict,
            source_name=source_name)




from monosi.events import track_event
from monosi.tasks.base import TaskBase
from monosi.tasks.run import RunnerTask

from monosi.scheduler.job import MonitorJob
from monosi.scheduler.manager import JobManager

class ScheduleMonitorsTask(TaskBase):
    def __init__(self, args, config, manager, monitors):
        super.__init__(args, config)
        self.manager = manager
        self.monitors = monitors

    def run(self, *args, **kwargs):
        track_event(self.config, action="scheduling_monitors")
        for monitor in self.monitors:
            job = RunnerTask(self.args, self.config, [monitor])
            self.manager.add_job(job, self.args, minutes=monitor.schedule.minutes)




import abc
from typing import List, Optional

from monosi.config.configuration import Configuration
from monosi.project import Project

class TaskBase:
    def __init__(self, args, config):
        self.args = args
        self.config = config

    @classmethod
    def from_args(cls, args):
        try:
            config = Configuration.from_args(args)
        except:
            raise Exception("There was an issue creating the task from args.")

        return cls(args, config)

    @classmethod
    def run_task(cls, *args, **kwargs):
        task = cls(*args)
        return task.run(*args, **kwargs)

    @abc.abstractmethod
    def run(self, *args, **kwargs):
        raise NotImplementedError('Implementation for task does not exist.')

class ProjectTask(TaskBase):
    def __init__(self, args, config):
        super().__init__(args, config)
        self.project: Optional[Project] = None
        self.task_queue: List[TaskBase] = []

    def load_project(self):
        self.project = Project.from_configuration(self.config)
        self.task_queue = self._create_tasks()

    def _initialize(self):
        self.load_project()

    @abc.abstractmethod
    def _create_tasks(self):
        raise NotImplementedError

    @abc.abstractmethod
    def _process_tasks(self):
        raise NotImplementedError

    def run(self):
        self._initialize()
        return self._process_tasks()




import abc

from monosi.config.configuration import Configuration
import monosi.utils.yaml as yaml

class Parser:
    def __init__(self, configuration: Configuration):
        self.configuration = configuration

    @abc.abstractmethod
    def parse_file(self, file, project):
        pass

    @property
    def default_schema(self):
        schema = self.configuration.config.schema
        if schema is None:
            raise Exception("Could not resolve table name without default schema or table name in format DATABASE.SCHEMA.TABLE")

        return schema

    @property
    def default_database(self):
        database = self.configuration.config.database
        if database is None:
            raise Exception("Could not resolve table name without default database or table name in format DATABASE.SCHEMA.TABLE")

        return database

class YamlParser(Parser):
    def __init__(self, configuration, keys):
        super().__init__(configuration)
        self.keys = keys

    def extract_from_file(self, file):
        filepath = file.filepath
        file_dict = yaml.parse_yaml(filepath)

        # TODO: Implement YAML Search
        parser_contents = file_dict
        for key in self.keys:
            parser_contents = parser_contents[key]

        return parser_contents




import abc
from dataclasses import dataclass
from typing import List, Optional, Type

from .column import Column
from .dialect import Dialect

@dataclass
class DriverConfig:
    schema: Optional[str] = None
    database: Optional[str] = None

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'retrieve_data') and 
                callable(subclass.retrieve_data) and 
                hasattr(subclass, 'validate') and 
                callable(subclass.validate) or 
                hasattr(subclass, 'from_dict') and 
                callable(subclass.from_dict) or 
                NotImplemented)

    @abc.abstractmethod
    def driver_name(self):
        raise NotImplementedError

    @abc.abstractmethod
    def retrieve_data(self, source_dict):
        raise NotImplementedError

    @abc.abstractmethod
    def validate(self, config_dict):
        raise NotImplementedError

    @abc.abstractmethod
    def to_dict(self):
        raise NotImplementedError

    @abc.abstractmethod
    def from_dict(self, config_dict):
        raise NotImplementedError

class BaseDriver:
    dialect: Type[Dialect]

    @abc.abstractmethod
    def test_connection(self):
        raise NotImplementedError

    @abc.abstractmethod
    def describe_table(self, table):
        raise NotImplementedError

    @abc.abstractmethod
    def execute_sql(self, sql):
        raise NotImplementedError

    @abc.abstractclassmethod
    def get_compiler(cls):
        raise NotImplementedError

    def _create_columns(self, columns_and_types) -> List[Column]:
        column_names = []
        column_count = {}

        columns = []

        for column_and_type in columns_and_types:
            name = column_and_type[0]
            column_type = column_and_type[1]

            try:
                curr_count = column_count[name] + 1
            except:
                curr_count = 0
            column_count[name] = curr_count

            column_name = "{}{}".format(name, (column_count[name] if curr_count else ''))
            column_names.append(column_name)

            columns.append(Column(column_name, column_type))

        return columns




STATUS_SCHEDULED = 0
STATUS_RUNNING = 1
STATUS_STOPPING = 2
STATUS_STOPPED = 3
STATUS_FAILED = 4
STATUS_SUCCEEDED = 5
STATUS_TIMEOUT = 6
STATUS_SCHEDULED_ERROR = 7

STATUS_DICT = {
    STATUS_SCHEDULED: 'scheduled',
    STATUS_RUNNING: 'running',
    STATUS_STOPPING: 'stopping',
    STATUS_STOPPED: 'stopped',
    STATUS_FAILED: 'failed',
    STATUS_SUCCEEDED: 'succeeded',
    STATUS_TIMEOUT: 'timeout',
    STATUS_SCHEDULED_ERROR: 'scheduled error'
}


import os
basedir = os.path.abspath(os.path.dirname(__file__))

monosi_path = os.path.expanduser("~/.monosi")
if not os.path.exists(monosi_path):
    os.makedirs(monosi_path)

class BaseConfig:
    """Base configuration."""
    SECRET_KEY = os.getenv('SECRET_KEY', 'my_precious')
    DEBUG = False
    SQLALCHEMY_DATABASE_URI = "sqlite:///" + monosi_path + "/sqlite.db"

class DevelopmentConfig(BaseConfig):
    """Development configuration."""
    DEBUG = True

class TestingConfig(BaseConfig):
    """Testing configuration."""
    DEBUG = True
    TESTING = True
    PRESERVE_CONTEXT_ON_EXCEPTION = False

class ProductionConfig(BaseConfig):
    """Production configuration."""
    DEBUG = False

Config = DevelopmentConfig
if os.getenv('FLASK_ENV') == "production":
    Config = ProductionConfig



from enum import Enum
from sys import stdout

class Color(Enum): # You may need to change color settings
    RED = '\033[31m'
    GREEN = '\033[32m'
    YELLOW = '\033[33m'
    BLUE = '\033[34m'
    ENDC = '\033[m'

class BaseFormatter:
    def __init__(self, output=stdout):
        self.output = output

    # def start(self, notification):
    #     self.start_sync_output()

    def write(self, text: str, color: Color):
        self.output.write(color.value)
        self.output.write(text)

    def stop(self, none_obj):
        pass



from monosi.drivers.dialect import GenericDialect

class SnowflakeDialect(GenericDialect):
    @classmethod
    def text_int_rate(cls):
        return "SUM(IFF(REGEXP_COUNT(TO_VARCHAR({}), '^([-+]?[0-9]+)$', 1, 'i') != 0, 1, 0)) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def text_number_rate(cls):
        return "SUM(IFF(REGEXP_COUNT(TO_VARCHAR({}), '^([-+]?[0-9]*[.]?[0-9]+([eE][-+]?[0-9]+)?)$', 1, 'i') != 0, 1, 0)) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def text_uuid_rate(cls):
        return "SUM(IFF(REGEXP_COUNT(TO_VARCHAR({}), '^([0-9a-fA-F]{{8}}-[0-9a-fA-F]{{4}}-[0-9a-fA-F]{{4}}-[0-9a-fA-F]{{4}}-[0-9a-fA-F]{{12}})$', 1, 'i') != 0, 1, 0)) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def text_all_spaces_rate(cls):
        return "SUM(IFF(REGEXP_COUNT(TO_VARCHAR({}), '^(\\\\s+)$', 1, 'i') != 0, 1, 0)) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def text_null_keyword_rate(cls):
        return "SUM(IFF(UPPER({}) IN ('NULL', 'NONE', 'NIL', 'NOTHING'), 1, 0)) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def zero_rate(cls):
        return "SUM(IFF(UPPER({}) IN ('NULL', 'NONE', 'NIL', 'NOTHING'), 1, 0)) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def negative_rate(cls):
        return "SUM(IFF({} < 0, 1, 0)) / CAST(COUNT(*) AS NUMERIC)"

    @classmethod
    def completeness(cls):
        return "COUNT({}) / CAST(COUNT(*) AS NUMERIC)"




import re
import os
from typing import Any, Dict
import yaml

def parse_yaml(path=None, data=None, tag='!ENV'):
    """
    Load a yaml configuration file and resolve any environment variables
    The environment variables must have !ENV before them and be in this format
    to be parsed: ${VAR_NAME}.
    E.g.:
    database:
        host: !ENV ${HOST}
        port: !ENV ${PORT}
    app:
        log_path: !ENV '/var/${LOG_PATH}'
        something_else: !ENV '${AWESOME_ENV_VAR}/var/${A_SECOND_AWESOME_VAR}'
    :param str path: the path to the yaml file
    :param str data: the yaml data itself as a stream
    :param str tag: the tag to look for
    :return: the dict configuration
    :rtype: dict[str, T]
    """
    # pattern for global vars: look for ${word}
    pattern = re.compile('.*?\${(\w+)}.*?')
    loader = yaml.SafeLoader

    # the tag will be used to mark where to start searching for the pattern
    # e.g. somekey: !ENV somestring${MYENVVAR}blah blah blah
    loader.add_implicit_resolver(tag, pattern, None)

    def constructor_env_variables(loader, node):
        """
        Extracts the environment variable from the node's value
        :param yaml.Loader loader: the yaml loader
        :param node: the current node in the yaml
        :return: the parsed string that contains the value of the environment
        variable
        """
        value = loader.construct_scalar(node)
        match = pattern.findall(value)  # to find all env variables in line
        if match:
            full_value = value
            for g in match:
                full_value = full_value.replace(
                    f'${{{g}}}', os.environ.get(g, g)
                )
            return full_value
        return value

    loader.add_constructor(tag, constructor_env_variables)

    if path:
        with open(path) as conf_data:
            return yaml.load(conf_data, Loader=loader)
    elif data:
        return yaml.load(data, Loader=loader)
    else:
        raise ValueError('Either a path or data should be defined as input')

def parse_file(path: str) -> Dict[str, Any]:
    if not os.path.isfile(path):
        raise Exception(
            'File not found: expected at path {}'
            .format(path))

    loaded_dict = parse_yaml(path)

    if not isinstance(loaded_dict, dict):
        raise Exception(
            'File does not parse to dictionary: {}'
            .format(path))

    return loaded_dict

def write_file(path: str, contents):
    with open(path, 'w') as f:
        yaml.dump(contents, f, default_flow_style=False, sort_keys=False)




from dataclasses import dataclass
from enum import Enum

class MetricType(Enum):
    CUSTOM = 'custom'

@dataclass
class MetricBase:
    type: MetricType

    def alias(self):
        return self.type._value_




from monosi.scheduler.db import db

class Execution(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    job_id = db.Column(db.Integer, nullable=False)
    status = db.Column(db.Integer, nullable=False)
    result = db.Column(db.Text, nullable=True)
    created_at = db.Column(db.DateTime(timezone=True), nullable=False, default=db.func.now())
    updated_at = db.Column(db.DateTime(timezone=True), nullable=False, default=db.func.now(), onupdate=db.func.now())

    __tablename__ = "executions"

    @classmethod
    def get_by_id(cls, _id):
        return cls.query.filter(cls.id == _id).one()
        
    def create(self):
        try:
            db.session.add(self)
            db.session.commit()
        except Exception as e:
            print(e)
            db.session.rollback()

    def update(self, updates):
        for k, v in updates.items():
            setattr(self, k, v)
        db.session.commit()

    def delete(self):
        db.session.delete(self)
        db.session.commit()

    @classmethod
    def all(cls):
        return cls.query.all()



from monosi.runner import Runner
from .base import ProjectTask, TaskBase

class RunnerTask(TaskBase):
    def __init__(self, args, config, monitors):
        super().__init__(args, config)
        self.monitors = monitors

    def run(self, *args, **kwargs):
        runner = Runner(self.config, self.monitors)
        runner.run()

class RunMonitorsTask(ProjectTask):
    def _create_tasks(self):
        if self.project is None:
            raise Exception("Project was not loaded before running monitors.")

        return [RunnerTask(self.args, self.config, self.project.monitors)]

    def _process_tasks(self):
        results = [task.run() for task in self.task_queue]
        return results



from .base import Color
from .text import TextFormatter

class ProgressFormatter(TextFormatter):
    def monitor_started(self, monitor):
        self.write(monitor.info(), Color.ENDC)
        self.write('\n\n', Color.ENDC)

    def monitor_finished(self, monitor):
        # self.write('\n\nMonitor stopped\n', Color.ENDC)
        pass

    def dump_skipped(self, skipped_tests):
        pass

    def test_started(self, none_obj):
        self.write('', Color.ENDC)

    def test_finished(self, none_obj):
        self.write('', Color.ENDC)

    def test_passed(self, none_obj):
        self.write('.', Color.GREEN)
        
    def test_pending(self, none_obj):
        self.write('*', Color.YELLOW)

    def test_failed(self, none_obj):
        self.write('F', Color.RED)

    def start_dump(self, none_obj):
        # self.write('start_dump', Color.ENDC)
        pass



# -*- coding: utf-8 -*-

from os import path
import pkg_resources
from setuptools import setup, find_namespace_packages

VERSION = '0.0.2.post1'

with open('README.md') as f:
    readme = f.read()

with open('LICENSE') as f:
    license = f.read()

install_requires = []
with open(path.abspath("requirements.txt"), "r") as f:
    requirements_txt = f.readlines()
    install_requires = [
        str(requirement)
        for requirement
        in pkg_resources.parse_requirements(requirements_txt)
    ]

setup(
    name='monosi',
    version=VERSION,
    description='Monosi - Data observability & monitoring toolkit',
    # long_description=readme,
    author='Vocable Inc.',
    author_email='support@monosi.dev',
    url='https://github.com/monosidev/monosi',
    license=license,
    install_requires=install_requires,
    packages=find_namespace_packages(include=['monosi', 'monosi.*'], exclude=('tests', 'docs')),
    entry_points = {
        'console_scripts': [
            'monosi=monosi.__main__:main',
        ],
    },
)



from dataclasses import dataclass, field
from typing import List, Type
import snowflake.connector
from monosi.config.configuration import Configuration
from monosi.events import track_event

from monosi.drivers import BaseDriver, DriverConfig
from monosi.drivers.column import Column, ColumnDataType

from .dialect import SnowflakeDialect

@dataclass
class SnowflakeConfig(DriverConfig):
    account: str = field(default_factory=str)
    password: str = field(default_factory=str)
    user: str = field(default_factory=str)
    warehouse: str = field(default_factory=str)
    host: str = field(default_factory=str)

    def driver_name(self):
        return "snowflake"

    @classmethod
    def retrieve_data(cls, source_dict):
        return {
            'user': source_dict.get('user'),
            'password': source_dict.get('password'),
            'database': source_dict.get('database'),
            'account': source_dict.get('account'),
            'host': cls._host(source_dict.get('account')),
            'warehouse': source_dict.get('warehouse'),
        }

    @classmethod
    def validate(cls, config_dict):
        pass

    @classmethod
    def from_dict(cls, config_dict):
        return cls(**config_dict)
    
    @classmethod
    def _host(cls, account, region=None):
        if region:
            return "{}.{}.snowflakecomputing.com".format(account, region)

        return "{}.snowflakecomputing.com".format(account)
 
    def to_dict(self):
        return {
            'user': self.user,
            'password': self.password,
            'database': self.database,
            'account': self.account,
            'host': self.__class__._host(self.account),
            'warehouse': self.warehouse,
        }

SNOWFLAKE_TYPES = {
    0: ColumnDataType.INTEGER,
    1: ColumnDataType.FLOAT,
    2: ColumnDataType.STRING,
    3: ColumnDataType.DATE,
    4: ColumnDataType.DATETIME,
    5: ColumnDataType.STRING,
    6: ColumnDataType.DATETIME,
    7: ColumnDataType.DATETIME,
    8: ColumnDataType.DATETIME,
    13: ColumnDataType.BOOLEAN,
}

def system_tables():
    return ['databases', 'external_tables', 'file_formats', 'functions', 'load_history', 'object_privileges', 'table_privileges', 'pipes', 'procedures', 'referential_constraints', 'replication_databases', 'schemata', 'sequences', 'stages', 'table_constraints', 'table_storage_metrics', 'tables', 'usage_privileges', 'views']

def resolve_to_type_from_str(type_str):
    type_str = type_str.lower()
    if 'varchar' in type_str:
        return ColumnDataType.STRING
    elif 'number' in type_str:
        return ColumnDataType.INTEGER
    elif 'date' in type_str:
        return ColumnDataType.DATE

    return ColumnDataType.STRING

@dataclass
class SnowflakeConfiguration(Configuration):
    config: SnowflakeConfig

class SnowflakeDriver(BaseDriver):
    def __init__(self, config: SnowflakeConfiguration):
        self.config = config.config
        self.dialect = SnowflakeDialect
        try:
            connection_details = self.config.to_dict()
            self._instance = snowflake.connector.connect(
                **connection_details
            )
            track_event(config, action="database_connection_success", label="snowflake")
        except Exception as e:
            track_event(config, action="database_connection_fail", label="snowflake")

    def _before_execute(self, cs):
        if self.config.warehouse is not None: 
            warehouse_sql = "USE WAREHOUSE {}".format(self.config.warehouse) # DANGER
            cs.execute(warehouse_sql)
        if self.config.database is not None: # DANGER
            database_sql = "USE {}".format(self.config.database)
            cs.execute(database_sql)

    def _retrieve_type(self, data_type, scale):
        resolved_type = SNOWFLAKE_TYPES.get(data_type, None)
        if resolved_type == ColumnDataType.STRING and scale is not None and scale > 0:
            resolved_type = ColumnDataType.FLOAT

        return resolved_type

    def _retrieve_results(self, cs):
        column_and_types = [(d[0], self._retrieve_type(d[1], d[5])) for d in cs.description]
        columns = self._create_columns(column_and_types)

        rows = [dict(zip([column.name for column in columns], row)) for row in cs]

        results = {"columns": columns, "rows": rows}
        return results

    def describe_table(self, table) -> List[Column]:
        describe_table_sql = "describe table {}".format(table) # DANGER
        results = self.execute_sql(describe_table_sql)
        
        columns = []
        for row in results['rows']:
            if row['kind'] == 'COLUMN':
                column = Column(row['name'], resolve_to_type_from_str(row['type']))
                columns.append(column)

        return columns

    def execute_sql(self, sql, **params):
        cs = self._instance.cursor()
        results = []
        try:
            self._before_execute(cs)

            cs.execute(sql, params)
            results = self._retrieve_results(cs)
        finally:
            cs.close()

        return results
 
    def metadata(self):
        DATABASE_METADATA_SQL = """
        SELECT
            lower(c.table_name) AS name,
            lower(c.column_name) AS col_name,
            lower(c.data_type) AS col_type,
            c.comment AS col_description,
            lower(c.ordinal_position) AS col_sort_order,
            lower(c.table_catalog) AS database,
            lower(c.table_schema) AS schema,
            t.comment AS description,
            decode(lower(t.table_type), 'view', 'true', 'false') AS is_view
        FROM
            {database}.INFORMATION_SCHEMA.COLUMNS AS c
        LEFT JOIN
            {database}.INFORMATION_SCHEMA.TABLES t
                ON c.TABLE_NAME = t.TABLE_NAME
                AND c.TABLE_SCHEMA = t.TABLE_SCHEMA;
        """.format(database=self.config.database)
        results = self.execute_sql(DATABASE_METADATA_SQL)
        filtered_results = {}
        filtered_results['columns'] = results['columns']
        filtered_results['rows'] = []
        rows = results['rows']
        for row in rows:
            if row['NAME'].lower() not in system_tables():
                filtered_results['rows'].append(row)

        return filtered_results





from monosi.scheduler.base import MonosiScheduler
from monosi.scheduler.db import db

class JobManager:
    singleton = None

    def __init__(self, app=None):
        if JobManager.singleton is None:
            self.app = app
            self.scheduler = MonosiScheduler()
            if app is not None:
                self.init_app(app)
            JobManager.singleton = self

    def init_app(self, app):
        self.scheduler.init_app(app)

        db.init_app(app)
        with app.app_context():
            db.create_all()

        self.start()

    def start(self):
        return self.scheduler.start()

    def stop(self):
        return self.scheduler.shutdown()

    def add_job(self, job, args=None, trigger='interval', minutes=720, **kwargs):
        return self.scheduler.add_scheduler_job(job, args=args, trigger=trigger, minutes=minutes)

    def pause_job(self, job_id):
        return self.scheduler.pause_job(job_id)

    def get_job(self, job_id):
        return self.scheduler.get_job(job_id)

    def get_jobs(self):
        return self.scheduler.get_jobs()

    def remove_job(self, job_id):
        return self.scheduler.remove_job(job_id)

    def resume_job(self, job_id):
        return self.scheduler.resume_job(job_id)




from flask_apscheduler import APScheduler
import json
import uuid
import logging

from monosi.scheduler.db import db
from monosi.scheduler.models.execution import Execution
import monosi.scheduler.constants as constants

logging.getLogger('apscheduler').setLevel(logging.DEBUG)

def import_from_path(path):
    components = path.split('.')
    module = __import__('.'.join(components[:-1]))
    for comp in components[1:-1]:
        module = getattr(module, comp)
    return getattr(module, components[-1])

class MonosiScheduler(APScheduler):
    @classmethod
    def run_job(cls, job_class_str, job_id, *args, **kwargs):
        execution = Execution(
            job_id=job_id,
            state=constants.EXECUTION_STATUS_SCHEDULED
        )
        execution.create()
        execution_id = execution.id

        try:
            job_class = import_from_path(job_class_str)
            execution.update(updates={"state": constants.EXECUTION_STATUS_SCHEDULED})

            cls.run_scheduler_job(job_class, execution_id, *args, **kwargs)
        except Exception:
            execution.update(updates={"state": constants.EXECUTION_STATUS_SCHEDULED_ERROR})

            return None
        return execution_id

    @classmethod
    def run_scheduler_job(cls, job_class, execution_id, *args, **kwargs):
        try:
            execution = Execution.get_by_id(execution_id)
            execution.update(updates={"state": constants.EXECUTION_STATUS_RUNNING})

            result = job_class.run()
            result_json = json.dumps(result, indent=4, sort_keys=True)

            execution.update(updates={
                "state": constants.EXECUTION_STATUS_SUCCEEDED,
                "result": result_json
            })

        except Exception as err:
            print("Error: {0}".format(err))
            if execution:
                execution.update({
                    "state": constants.EXECUTION_STATUS_FAILED
                })

    def add_scheduler_job(self, job_class_str, args=None, trigger='interval', minutes=720, **kwargs):
        job_id = uuid.uuid4().hex

        if not args:
            args = []
        arguments = [job_class_str, job_id]
        arguments.extend(args)

        self.add_job(func=self.run_job, id=job_id, args=arguments, trigger=trigger, minutes=minutes)

        return job_id



from dataclasses import dataclass
from enum import Enum

class ColumnDataType(Enum):
    STRING = 'string'
    DATE = 'date'
    DATETIME = 'datetime'
    INTEGER = 'integer'
    FLOAT = 'float'
    BOOLEAN = 'boolean'

@dataclass
class Column:
    name: str
    data_type: ColumnDataType
